/*
 * win32_process.h
 *
 * Copyright (C) 2006  Insigme Co., Ltd
 *
 * This software has been developed while working on the Linux Unified Kernel
 * project (http://www.longene.org) in the Insigma Research Institute,  
 * which is a subdivision of Insigma Co., Ltd (http://www.insigma.com.cn).
 * 
 * The project is sponsored by Insigma Co., Ltd.
 *
 * The authors can be reached at linux@insigma.com.cn.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of  the GNU General  Public License as published by the
 * Free Software Foundation; either version 2 of the  License, or (at your
 * option) any later version.
 *
 * Revision History:
 *   Jan 2006 - Created.
 */

#ifndef _WIN32_PROCESS_H
#define _WIN32_PROCESS_H

#include <linux/fs.h>
#include <linux/list.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/nls.h>
#include <linux/win32_thread.h>
#include <linux/winternl.h>

#ifdef CONFIG_UNIFIED_KERNEL
unsigned int get_current_time(void);
#define current_time get_current_time() 

/* Process state */
#define PROCESS_STATE_TERMINATED	1
#define PROCESS_STATE_ACTIVE		2

/* Process priority classes */
#define PROCESS_PRIORITY_CLASS_HIGH	(4) /* FIXME */
#define PROCESS_PRIORITY_CLASS_IDLE	(0) /* FIXME */
#define PROCESS_PRIORITY_CLASS_NORMAL	(2) /* FIXME */
#define PROCESS_PRIORITY_CLASS_REALTIME	(5) /* FIXME */
#define PROCESS_PRIORITY_CLASS_BELOW_NORMAL (1) /* FIXME */
#define PROCESS_PRIORITY_CLASS_ABOVE_NORMAL (3) /* FIXME */

static __inline__ struct eprocess *get_eprocess(struct ethread *thread)
{
	struct eprocess *ep = thread->threads_process;
	
	return ep;
} /* end get_eprocess() */

/*
 * kprocess
 */
struct kprocess
{
	struct dispatcher_header	header;		   	/* 000 */
	struct list_head		profile_list_head;	/* 010 */
	physical_address_t		directory_table_base;	/* 018 */
#if defined(_M_IX86)
	struct kgd_entry		ldt_descriptor;		/* 020 */
	struct kidt_entry		Int21_descriptor;  	/* 028 */
	unsigned short			iopm_offset;		/* 030 */
	unsigned char			iopl;		  	/* 032 */
	unsigned char			unused;			/* 033 */
#endif
	unsigned long			active_processors;  	/* 034 */
	unsigned long			kernel_time;		/* 038 */
	unsigned long			user_time;	  	/* 03C */
	struct list_head		ready_list_head;	/* 040 */
	struct list_head		swap_list_entry;	/* 048 */
	void*				vdm_trapc_handler; 	/* 04C */
	struct list_head		thread_list_head;	/* 050 */
	spinlock_t			process_lock;	  	/* 068 */
	unsigned long			affinity;	  	/* 06C */
	union
	{
		struct
		{
			unsigned long	auto_alignment:1;	/* 070.0 */
			unsigned long	disable_boost:1;	/* 070.1 */
			unsigned long	disable_quantum:1;	/* 070.2 */
			unsigned long	reserved_flags:29;	/* 070.3 */
		};
		unsigned long		process_flags;		/* 070 */
	};
	char				base_priority;		/* 074 */
	char				quantum_reset;		/* 075 */
	unsigned char			state;		 	/* 076 */
	unsigned char			thread_seed;		/* 077 */
	unsigned char			power_state;		/* 078 */
	unsigned char			ideal_node;	 	/* 079 */
	unsigned char			visited;	   	/* 07A */
	struct kexecute_options		flags;		 	/* 07B */
	unsigned long			stack_count;		/* 07C */
	struct list_head		process_list_entry; 	/* 080 */
};

struct eprocess
{
	struct kprocess			pcb;			 	/* 000 */
	unsigned long			exit_status;		  	/* 088 */
	spinlock_t			process_lock;		 	/* 08C */
	struct kevent			lock_event;			/* 0A4 */
	unsigned long			lock_count;			/* */
	struct kthread			*lock_owner;		   	/* */
	large_integer_t			create_time;		   	/* */
	large_integer_t			exit_time;		 	/* */
	void 				*rundown_protect;		/* */
	void*				unique_processid;		/* */
	struct list_head		active_process_links;		/* */
	
	/* quota fields */
	unsigned long			quota_usage[3];			/* */
	unsigned long			quota_peak[3];			/* */
	unsigned long			commit_charge;		 	/* */
	
	/* vm */
	unsigned long			peak_virtual_size;		/* */
	unsigned long			virtual_size;		  	/* */
	
	struct mmsupport		vm;			   	/* */
	struct list_head		session_process_links;		/* */
	
	void*				debug_port;			/* */
	void*				exception_port;			/* */
	struct handle_table*		object_table;		  	/* */
	
	/* security */
	struct ex_fast_ref		token;				/* */
	
	unsigned long			working_set_page;		/* */
	struct kguarded_mutex		address_creation_lock;		/* */
	spinlock_t			hyper_spacelock;	   	/* */
	struct ethread*			fork_in_progress;		/* */
	unsigned long			hardware_trigger;	  	/* */
	void*				pae_top;		   	/* */
	unsigned long			modified_page_count;		/* */
	struct mm_avl_table		vad_root;		  	/* */
	void*				vad_free_hint;			/* */
	void*				clone_root;			/* */
		
	struct mm_avl_table		physical_vadroot;	  	/* */
	
	unsigned long			number_of_private_pages;	/* */
	unsigned long			number_of_locked_pages;		/* */
	unsigned short			next_page_color;		/* */
	
	/* used by debug subsystem */
	void*				section_object;			/* */
	
	/* peb */
	void*			peb;			 	/* */
	void*				section_base_address;		/* */
	
	struct eprocess_quota_block*	quota_block;		  	/* */
	unsigned long			last_thread_exit_status;	/* */
	
	struct pagefault_history* 	working_set_watch;		/* */
	void*				win32_window_station;		/* */
	void *				inherited_from_unique_pid;	/* */
	access_mask_t			granted_access;			/* */
	unsigned long			def_hard_error_processing; 	/* */
	void*				ldt_information;	   	/* */
	void*				vdm_objects;		 	/* */
	void*				device_map;			/* */
	
	void*				spare0[3];			/* */
	
	union
	{
		struct hardware_pte_x86	pagedirectory_pte;	 	/* */
		unsigned long long	filler;				/* */
	};
	char				image_file_name[16];		/* */
	unsigned char			priority_class;			/* */
	union
	{
		struct
		{
			unsigned char	subsystem_minor_version;	/* */
			unsigned char	subsystem_major_version;	/* */
		};
		unsigned short		subsystem_version;		/* */
	};
	struct w32process* 		win32process;
	struct ejob*			job;			 	/* */
	unsigned long			job_status;			/* */
	struct list_head		job_links;		 	/* */
	void*				locked_pages_list;		/* */
	
	/* used by rdr/security for authentication */
	void*				security_port;		 	/* */
	
	large_integer_t			read_operation_count;		/* */
	large_integer_t			write_operation_count;		/* */
	large_integer_t			other_operation_count;		/* */
	large_integer_t			read_transfer_count;		/* */
	large_integer_t			write_transfer_count;		/* */
	large_integer_t			other_transfer_count;		/* */
	
	unsigned long			commit_charge_limit;		/* */
	unsigned long			commit_charge_peak;		/* */
	
	struct list_head		thread_list_head;		/* */
	
	union
	{
		struct
		{
			unsigned long	create_reported:1;
			unsigned long	no_debug_inherit:1;
			unsigned long	process_exiting:1;
			unsigned long	process_delete:1;
			unsigned long	wow64_split_pages:1;
			unsigned long	vm_deleted:1;
			unsigned long	outswap_enabled:1;
			unsigned long	outswapped:1;
			unsigned long	fork_failed:1;
			unsigned long	wow64_va_space4Gb:1;
			unsigned long	address_space_initialized:2;
			unsigned long	set_timer_resolution:1;
			unsigned long	break_on_termination:1;
			unsigned long	session_creation_underway:1;
			unsigned long	write_watch:1;
			unsigned long	process_in_session:1;
			unsigned long	override_address_space:1;
			unsigned long	has_address_space:1;
			unsigned long	launch_prefetched:1;
			unsigned long	inject_inpage_errors:1;
			unsigned long	vm_top_down:1;
			unsigned long	image_notify_done:1;
			unsigned long	pde_update_needed:1;
			unsigned long	vdm_allowed:1;
			unsigned long	smap_allowed:1;
			unsigned long	create_failed:1;
			unsigned long	default_io_priority:3;
			unsigned long	spare1:1;
			unsigned long	spare2:1;
		};
		unsigned long		flags;					/* */
	};

	unsigned long			cookie;				   	/* */
	unsigned long			session;			  	/* */
	unsigned long			active_threads;			   	/* */
	struct ex_fast_ref		prefetch_trace;				/* */
	void*				awe_info;				/* */
	struct se_audit_proc_creation_info	se_audit_proc_creation_info;	/* */
	struct list_head			mm_process_links;		/* */

	struct nls_table*		ep_nls;	/* unicode-ascii translation */
	rwlock_t			ep_lock;
	struct list_head		ep_reserved_head;
	struct list_head		ep_mapped_head;
    void                    *ep_handle_info_table;

	/*for NtNotifyDirectoryChange */
	spinlock_t                      watch_lock;
	long				watch_fd;
	pid_t				watch_thread;
}; /* struc eprocess */

typedef struct eprocess EPROCESS, *PEPROCESS;

#endif /* CONFIG_UNIFIED_KERNEL */
#endif /* _WIN32_PROCESS_H */
